@@ -4,7 +4,7 @@ use v5.014;
 use strict;
 use warnings;
 
-our $VERSION = '0.02';
+our $VERSION = '0.04';
 
 use JSON qw/ encode_json /;
 use Text::CSV;
@@ -99,9 +99,11 @@ sub csv {
     for my $index (sort keys %$data_matrix) {
         my $data= $data_matrix->{$index};
         my %record;
-        @record{@$index_column_names}= split /\0/, $index;
-        @record{keys %$data}= values %$data;
-
+        @record{@$index_column_names}= _deserialize_tuple($index);
+        for my $column_key (keys %$data) {
+            my $friendly_column_name= join "_", _deserialize_tuple($column_key);
+            $record{$friendly_column_name}= $data->{$column_key};
+        }
         @column_names{keys %record}= (1) x keys %record;
 
         push @records, \%record;
@@ -109,7 +111,7 @@ sub csv {
     my @column_names= sort keys %column_names;
     my $render_header= scalar grep $_, @column_names;
 
-    my $csv= Text::CSV->new();
+    my $csv= Text::CSV->new({binary => 1});
 
     my @result;
     if($render_header) {
@@ -133,6 +135,8 @@ sub csv {
     return join "\n", @result;
 }
 
+# utility function to iterate over key => value pairs with the added
+# bonus that it also works for arrays
 sub _foreach(&$) {
     my ($codeblock, $it)= @_;
 
@@ -154,6 +158,14 @@ sub _foreach(&$) {
     }
 }
 
+sub _serialize_tuple {
+    return pack "(S/a)*", @_;
+}
+
+sub _deserialize_tuple {
+    return unpack "(S/a)*", $_[0];
+}
+
 sub _recurse_pattern {
     my ($self, $cur_data, $pattern, $column_name_prefix, $index_prefix)= @_;
 
@@ -170,12 +182,10 @@ sub _recurse_pattern {
                 for my $key (@keys) {
                     my $recurse_data;
                     if(ref $cur_data eq 'HASH' && exists $cur_data->{$key}) {
-                        $recurse_data= $cur_data->{$key};
-                    } elsif(ref $cur_data eq 'ARRAY') {
-                        $recurse_data= $cur_data->[$key];
+                        _recurse_pattern($self, $cur_data->{$key}, \@p, [@$column_name_prefix, $key], $index_prefix)
+                    } elsif(ref $cur_data eq 'ARRAY' && exists $cur_data->[$key]) {
+                        _recurse_pattern($self, $cur_data->[$key], \@p, [@$column_name_prefix, $key], $index_prefix)
                     }
-                    _recurse_pattern($self, $recurse_data, \@p, [@$column_name_prefix, $key], $index_prefix)
-                        if $recurse_data;
                 }
             } elsif($p =~ /^<(.*)>$/) {
                 _foreach {
@@ -183,14 +193,11 @@ sub _recurse_pattern {
                     _recurse_pattern($self, $value, \@p, $column_name_prefix, [@$index_prefix, $key]);
                 } $cur_data;
             } else {
-                my $recurse_data;
                 if(ref $cur_data eq 'HASH' && exists $cur_data->{$p}) {
-                    $recurse_data= $cur_data->{$p};
-                } elsif(ref $cur_data eq 'ARRAY') {
-                    $recurse_data= $cur_data->[$p];
+                    _recurse_pattern($self, $cur_data->{$p}, \@p, $column_name_prefix, $index_prefix)
+                } elsif(ref $cur_data eq 'ARRAY' && exists $cur_data->[$p]) {
+                    _recurse_pattern($self, $cur_data->[$p], \@p, $column_name_prefix, $index_prefix)
                 }
-                _recurse_pattern($self, $recurse_data, \@p, $column_name_prefix, $index_prefix)
-                    if $recurse_data;
             }
             1;
         } or do {
@@ -202,10 +209,8 @@ sub _recurse_pattern {
         my $cell_value= ref $cur_data
                       ? encode_json($cur_data)
                       : $cur_data;
-        my $column_name= @$column_name_prefix
-                       ? join("_", @$column_name_prefix)
-                       : $self->{_default_column_name} || '';
-        $self->{data_matrix}{join("\0", @$index_prefix)}{$column_name}= $cell_value;
+        my @column_tuple= @$column_name_prefix ? @$column_name_prefix : ($self->{_default_column_name} || '');
+        $self->{data_matrix}{_serialize_tuple(@$index_prefix)}{_serialize_tuple(@column_tuple)}= $cell_value;
     }
 }
 
